package edu.mit.csail.pag.recrash;

import java.io.IOException;
import java.lang.instrument.ClassFileTransformer;
import java.lang.instrument.IllegalClassFormatException;
import java.lang.instrument.Instrumentation;
import java.security.ProtectionDomain;

import checkers.quals.Default;
import checkers.quals.NonNull;
import checkers.quals.Nullable;

import edu.mit.csail.pag.recrash.transformation.Transformer;

@Default("checkers.quals.NonNull")
public class Agent {

	/**
	 * Called when Java is invoked with -javaagent pointing to a jar with this
	 * class as premain agent.
	 * 
	 * @throws IOException
	 */
	public static void premain(@NonNull
	String agentArgs, Instrumentation inst) throws IOException {

		final Transformer trans = new Transformer();

		if (agentArgs != null) {
			@Nullable
			String[] args = agentArgs.split("(  *)|(, *)");
			trans.parseArgs(args, 0);
		}

		if (Logger.verbose) {
			System.out.format("In premain, agentargs ='%s', "
					+ "Instrumentation = '%s'\n", agentArgs, inst);
		}

		inst.addTransformer(new ClassFileTransformer() {

			public @Nullable
			byte[] transform(ClassLoader loader, String className,
					Class<?> classBeingRedefined,
					ProtectionDomain protectionDomain, byte[] classfileBuffer)
					throws IllegalClassFormatException {

				// return null if we don't have to instrument the given class
				if (!Util.shouldInstrumentThisClass(Util
						.transClassNameDotToSlash(className))) {
					return null;
				}

				try {
					// return instrumented class
					return trans.treeAPITransform(classfileBuffer);
				} catch (IOException e) {
					e.printStackTrace();
					return null;
				}
			}
		});
	}
}
