package edu.mit.csail.pag.recrash.tracer;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.zip.GZIPOutputStream;

import checkers.quals.*;

import edu.mit.csail.pag.recrash.Logger;
import edu.mit.csail.pag.recrash.Util;
import edu.mit.csail.pag.recrash.generator.CrashTestCaseGenerator;
import edu.mit.csail.pag.recrash.tracer.item.StackTraceItem;

/**
 * Utility class for TraceItem It handles adding/reading objects using XStream
 * 
 * @author hunkim
 * 
 */
/* @Default("checkers.quals.NonNull") */
public class TraceWriter {

	public static void setProgramNameVersionNumber(String nameAndNumber) {
		ThreadHandler.setProgramNameVersionNumber(nameAndNumber);
	}

	/**
	 * Write current trace (traceItem) with Exception information
	 * 
	 * @param crashedException
	 */
	public static/* @Nullable */File writeTrace(Throwable crashedException) {
		/* @Nullable */File traceFile = null;
		OutputStreamWriter fw;

		if (crashedException == null) {
			Logger.info("Nothing to write: Exception is null!");
			return null;
		}
		Logger.info("Writing trace information to file");

		StackTraceItem stackTraceItem = ThreadHandler.getCurrentItem();

		if (stackTraceItem == null) {
			Logger.error("Nothing write in this thread. Skip this!");
			return null;
		}

		stackTraceItem.setCrashedException(crashedException);

		try {
			traceFile = File.createTempFile("reCrash", ".trace.gz", new File(
					Util.getTmpDirectory()));
			assert (traceFile != null);

			Logger.info("Trace file:" + traceFile.getAbsolutePath());

			fw = new OutputStreamWriter(new GZIPOutputStream(
					new FileOutputStream(traceFile)));

			if (Util.MAX_STACK_TRACE_TO_WRITE != -1) {
				stackTraceItem = stackTraceItem
						.getReducedStackItem(Util.MAX_STACK_TRACE_TO_WRITE);
			}

			Serializer.storeObject(stackTraceItem, fw);
			Logger
					.info("Trace file:" + traceFile.getAbsolutePath()
							+ " done!!");
			fw.close();
		} catch (IOException e) {
			e.printStackTrace();
		}

		// FIXME: For convenience we hook, reCrash unit test generator
		// Should be separate process in the production
		try {
			if (traceFile != null) {
				/* @Nullable */String traceFileName = traceFile
						.getAbsolutePath();
				if (traceFileName != null) {
					CrashTestCaseGenerator.main(new String[] { traceFileName,
					// "/afs/csail.mit.edu/u/h/hunkim/workspace/reCrash/generated_tests"
							Util.getTmpDirectory() });
				}
			}
		} catch (Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		Logger.info("Trace file:" + traceFile.getAbsolutePath());
		return traceFile;
	}

	public static void addStatic(String fullyQualifiedStaticClassName) {
	}

	/**
	 * Add static class used in the method TODO: static field is not stored
	 * using XStream. We need to store all fields including static field
	 * 
	 * @param fullyQualifiedStaticClassName
	 */
	public static void addStaticToDev(String fullyQualifiedStaticClassName) {

		assert (fullyQualifiedStaticClassName != null);
		Object staticObject = null;

		try {
			staticObject = Class.forName(fullyQualifiedStaticClassName)
					.newInstance();
		} catch (Exception e) {
			if (Logger.verbose) {
				Logger.error("Static :" + fullyQualifiedStaticClassName
						+ " does not exist!");
			}
		}

		// If the object is null, don't save it
		if (staticObject == null) {
			return;
		}

		// trace item should not be null
		// or addObject must start the "this"
		StackTraceItem traceItem = ThreadHandler.getCurrentItem();
		assert (traceItem != null);

		String serializedXMLString = Serializer.toXMLString(staticObject);
		traceItem.addStatic(fullyQualifiedStaticClassName, serializedXMLString);
	}

	/**
	 * Start a new method and return nounce number
	 * 
	 * @param methodName
	 * @return
	 */

	public static int youMayCrash(String methodName, Object objects[],
			String types[], boolean doCloneObject[]) {
		if (Logger.verbose) {
			Logger.info(Thread.currentThread().getName() + " -> enter Method: "
					+ methodName + " arg nun: " + objects.length);
		}
		// Initialize trace item and reset MethodInfo
		return ThreadHandler.enterMethod(methodName, objects, types,
				doCloneObject);
	}

	public static void youAreOK(String methodName, int nounceNumber) {
		if (Logger.verbose) {
			Logger.info(Thread.currentThread().getName() + " <- exit Method: "
					+ methodName);
		}
		// Initialize trace item and reset MethodInfo
		ThreadHandler.exitMethod(methodName, nounceNumber);
	}
}
