package edu.mit.csail.pag.recrash.transformation;

import java.io.IOException;
import java.io.InputStream;

import org.objectweb.asm.ClassAdapter;
import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.Label;
import org.objectweb.asm.MethodVisitor;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.Type;
import org.objectweb.asm.commons.AdviceAdapter;
import edu.mit.csail.pag.recrash.Logger;
import edu.mit.csail.pag.recrash.Util;

/**
 * 
 * @author hunkim
 * @depricated
 */
public class AddTraceAdapter extends ClassAdapter {
	boolean findMain = false;
	boolean shouldInstrumentThis = false;
	String name = null;

	public AddTraceAdapter(ClassVisitor cv) {
		super(cv);
	}

	@Override
	public void visit(final int version, final int access, final String name,
			final String signature, final String superName,
			final String[] interfaces) {
		cv.visit(version, access, name, signature, superName, interfaces);
		this.name = name;
		shouldInstrumentThis = Util.shouldInstrumentThisClass(name);
	}

	@Override
	public MethodVisitor visitMethod(int access, String name, String desc,
			String signature, String[] exceptions) {
		if (!shouldInstrumentThis) {
			return cv.visitMethod(access, name, desc, signature, exceptions);
		}

		if (!findMain && "main".equals(name)
				&& "([Ljava/lang/String;)V".equals(desc)
				&& (access & Opcodes.ACC_PUBLIC) != 0
				&& (access & Opcodes.ACC_STATIC) != 0) {
			findMain = true;
			return visitMethod(access, Util.ORIGINAL_MAIN, desc, signature,
					exceptions);
		}

		MethodVisitor mv = cv.visitMethod(access, name, desc, signature,
				exceptions);

		if (mv == null) {
			return mv;
		}

		// transform the method
		mv = new AddTraceMethodAdapter(mv, access, name, desc);
		return mv;
	}

	@Override
	public void visitEnd() {
		if (shouldInstrumentThis && findMain) {
			MethodVisitor mv = visitMethod(Opcodes.ACC_PUBLIC
					+ Opcodes.ACC_STATIC, "main", "([Ljava/lang/String;)V",
					null, null);
			mv.visitCode();
			Label l0 = new Label();
			Label l1 = new Label();
			Label l2 = new Label();
			// mv.visitTryCatchBlock(l0, l1, l2, "java/lang/Throwable");
			mv.visitTryCatchBlock(l0, l1, l2, "java/lang/Exception");
			mv.visitLabel(l0);
			mv.visitLineNumber(67, l0);
			mv.visitVarInsn(Opcodes.ALOAD, 0);
			mv.visitMethodInsn(Opcodes.INVOKESTATIC, this.name,
					Util.ORIGINAL_MAIN, "([Ljava/lang/String;)V");
			mv.visitLabel(l1);
			Label l3 = new Label();
			mv.visitJumpInsn(Opcodes.GOTO, l3);
			mv.visitLabel(l2);
			mv.visitLineNumber(68, l2);
			mv.visitVarInsn(Opcodes.ASTORE, 1);
			Label l4 = new Label();
			mv.visitLabel(l4);
			mv.visitLineNumber(69, l4);
			mv.visitVarInsn(Opcodes.ALOAD, 1);
			mv.visitMethodInsn(Opcodes.INVOKESTATIC,
					"edu/mit/csail/pag/recrash/tracer/TraceWriter",
					"writeTrace", "(Ljava/lang/Throwable;)Ljava/io/File;");
			mv.visitInsn(Opcodes.POP);
			mv.visitLabel(l3);
			mv.visitLineNumber(71, l3);
			mv.visitInsn(Opcodes.RETURN);
			Label l5 = new Label();
			mv.visitLabel(l5);
			mv.visitLocalVariable("args", "[Ljava/lang/String;", null, l0, l5,
					0);
			mv
					.visitLocalVariable("e", "Ljava/lang/Throwable;", null, l4,
							l3, 1);
			mv.visitMaxs(1, 2);
			mv.visitEnd();
		}
		cv.visitEnd();
	}
}

/**
 * Method transform adapter
 * 
 * @author hunkim
 * 
 */
class AddTraceMethodAdapter extends AdviceAdapter {
	private String name;
	private String desc;
	private boolean instrumentThisMethod = true;
	private int nounceNumber;

	public AddTraceMethodAdapter(MethodVisitor mv, int access, String name,
			String desc) {
		super(mv, access, name, desc);
		if ("<init>".equals(name) || "<clint>".equals(name)
				|| (access & ACC_STATIC) != 0 // static
				|| (access & ACC_ABSTRACT) != 0 // no abstract
				|| (access & ACC_PUBLIC) == 0) { // not public
			instrumentThisMethod = false;
		}
		this.name = name;
		this.desc = desc;
	}

	private void addObjectInstrument(Type type, int argOrder) {
		assert (type != null);
		if (type.equals(Type.BOOLEAN_TYPE)) {
			mv.visitVarInsn(ILOAD, argOrder);
			mv.visitMethodInsn(INVOKESTATIC, "java/lang/Boolean", "valueOf",
					"(Z)Ljava/lang/Boolean;");
		} else if (type.equals(Type.CHAR_TYPE)) {
			mv.visitVarInsn(ILOAD, argOrder);
			mv.visitMethodInsn(INVOKESTATIC, "java/lang/Character", "valueOf",
					"(C)Ljava/lang/Character;");
		} else if (type.equals(Type.BYTE_TYPE)) {
			mv.visitVarInsn(ILOAD, argOrder);
			mv.visitMethodInsn(INVOKESTATIC, "java/lang/Byte", "valueOf",
					"(B)Ljava/lang/Byte;");
		} else if (type.equals(Type.SHORT_TYPE)) {
			mv.visitVarInsn(ILOAD, argOrder);
			mv.visitMethodInsn(INVOKESTATIC, "java/lang/Short", "valueOf",
					"(S)Ljava/lang/Short;");
		} else if (type.equals(Type.INT_TYPE)) {
			mv.visitVarInsn(ILOAD, argOrder);
			mv.visitMethodInsn(INVOKESTATIC, "java/lang/Integer", "valueOf",
					"(I)Ljava/lang/Integer;");
		} else if (type.equals(Type.FLOAT_TYPE)) {
			mv.visitVarInsn(FLOAD, argOrder);
			mv.visitMethodInsn(INVOKESTATIC, "java/lang/Float", "valueOf",
					"(F)Ljava/lang/Float;");
		} else if (type.equals(Type.LONG_TYPE)) {
			mv.visitVarInsn(LLOAD, argOrder);
			mv.visitMethodInsn(INVOKESTATIC, "java/lang/Long", "valueOf",
					"(J)Ljava/lang/Long;");
		} else if (type.equals(Type.DOUBLE_TYPE)) {
			mv.visitVarInsn(DLOAD, argOrder);
			mv.visitMethodInsn(INVOKESTATIC, "java/lang/Double", "valueOf",
					"(D)Ljava/lang/Double;");
		} else {
			mv.visitVarInsn(ALOAD, argOrder);
		}

		mv.visitMethodInsn(INVOKESTATIC,
				"edu/mit/csail/pag/recrash/tracer/TraceWriter", "addObject",
				"(Ljava/lang/Object;)V");
	}

	@Override
	protected void onMethodEnter() {
		if (!instrumentThisMethod) {
			return;
		}

		nounceNumber = newLocal(Type.INT_TYPE);

		mv.visitLdcInsn(name);
		mv.visitMethodInsn(INVOKESTATIC,
				"edu/mit/csail/pag/recrash/tracer/TraceWriter", "youMayCrash",
				"(Ljava/lang/String;)I");
		mv.visitVarInsn(ISTORE, nounceNumber);

		// THis (receiver)
		addObjectInstrument(Type.getType(Object.class), 0);

		Type[] types = Type.getArgumentTypes(desc);
		int localVarIndex = 1;
		for (int t = 0; t < types.length; t++) {

			Logger.info("\tAdd 'addObject' for " + types[t]);
			addObjectInstrument(types[t], localVarIndex);

			// size of type can be different
			localVarIndex += types[t].getSize();
		}
	}

	@Override
	protected void onMethodExit(int opcode) {
		if (!instrumentThisMethod) {
			return;
		}

		mv.visitLdcInsn(name);
		mv.visitVarInsn(ILOAD, nounceNumber);
		mv.visitMethodInsn(INVOKESTATIC,
				"edu/mit/csail/pag/recrash/tracer/TraceWriter", "youAreOK",
				"(Ljava/lang/String;I)V");
	}

	/**
	 * Use of the core API transformation
	 * 
	 * @param classIn
	 * @return
	 * @throws IOException
	 */
	public byte[] coreAPItransform(InputStream classIn) throws IOException {
		ClassReader cr = new ClassReader(classIn);
		ClassWriter cw = new ClassWriter(cr, ClassWriter.COMPUTE_MAXS);
		AddTraceAdapter ata = new AddTraceAdapter(cw);
		cr.accept(ata, ClassReader.SKIP_FRAMES);

		return cw.toByteArray();
	}
}
