package edu.mit.csail.pag.recrash.transformation;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.jar.JarEntry;
import java.util.jar.JarFile;
import java.util.jar.JarOutputStream;

import org.objectweb.asm.ClassReader;
import org.objectweb.asm.ClassWriter;
import org.objectweb.asm.Opcodes;
import org.objectweb.asm.Type;
import org.objectweb.asm.tree.AbstractInsnNode;
import org.objectweb.asm.tree.ClassNode;
import org.objectweb.asm.tree.FieldInsnNode;
import org.objectweb.asm.tree.FieldNode;
import org.objectweb.asm.tree.InsnList;
import org.objectweb.asm.tree.InsnNode;
import org.objectweb.asm.tree.IntInsnNode;
import org.objectweb.asm.tree.JumpInsnNode;
import org.objectweb.asm.tree.LabelNode;
import org.objectweb.asm.tree.LdcInsnNode;
import org.objectweb.asm.tree.LocalVariableNode;
import org.objectweb.asm.tree.MethodInsnNode;
import org.objectweb.asm.tree.MethodNode;
import org.objectweb.asm.tree.TryCatchBlockNode;
import org.objectweb.asm.tree.TypeInsnNode;
import org.objectweb.asm.tree.VarInsnNode;
import org.objectweb.asm.util.CheckClassAdapter;

import checkers.quals.NonNull;

import edu.mit.csail.pag.recrash.GetClass;
import edu.mit.csail.pag.recrash.Logger;
import edu.mit.csail.pag.recrash.Util;

public class Transformer implements Opcodes {

	private static final String RECRASH_ANNOTATION = "reCrash with";

	private boolean transformMain = true;

	private int minOpCount = 5;

	private String subjectProgramName = null;

	private boolean useClone;

	// immutable information
	private Mutability mutability = null;

	// second change mode
	SecondChangeMode scMode = null;

	// Keep name of the clone-able class names
	private Set<String> myClonnableTypeSet = new LinkedHashSet<String>();

	private boolean useSerialization = false;

	private boolean useShallowClone = false;

	static class reCrashWritePoint {
		AbstractInsnNode location;
		int exceptionArgNo;

		public reCrashWritePoint(AbstractInsnNode location, int no) {
			this.location = location;
			exceptionArgNo = no;
		}
	}

	public void transformDir(File srcDir, File descDir) throws IOException {
		assert (srcDir != null);
		assert (descDir != null);

		if (!srcDir.exists()) {
			Logger.fatal(srcDir + " does not exist!");
			return;
		}

		if (srcDir.isFile()) {
			transformFile(srcDir, descDir);
			return;
		}

		File[] subFiles = srcDir.listFiles();
		if (subFiles == null) {
			return;
		}

		for (int i = 0; i < subFiles.length; i++) {
			String subFileName = subFiles[i].getAbsolutePath();
			String addedOnFileName = subFileName.substring(srcDir
					.getAbsolutePath().length());
			if (subFiles[i].isFile()) {
				File descFile = new File(descDir, addedOnFileName);
				Logger.info("Transformming : " + subFiles[i] + " -> "
						+ descFile);
				transformFile(subFiles[i], descFile);
			} else if (subFiles[i].isDirectory()) {
				transformDir(subFiles[i], new File(descDir, addedOnFileName));
			}
		}
	}

	private void transformFile(File in, File out) throws IOException {
		if (in == null || out == null) {
			return;
		}

		// create the directory for the out
		if (out.getParentFile() != null) {
			out.getParentFile().mkdirs();
		}

		// ### FIXME: is there more reliable way to check if it is a class file
		// or a jar file
		if (in.getName().endsWith(".class")) {
			transformClassFile(in, out);
		} else if (in.getName().endsWith(".jar")) {
			transformJarFile(in, out);
		} else {
			Util.copyFile(in, out);
		}
	}

	/**
	 * Transform a jar file
	 * 
	 * @param in
	 * @param out
	 * @throws IOException
	 */
	private void transformJarFile(File in, File out) throws IOException {
		// jar file
		JarFile inJar = new JarFile(in);

		// jar output stream
		JarOutputStream outJarStream = new JarOutputStream(
				new FileOutputStream(out));

		// get all jar entries
		Enumeration<JarEntry> entries = inJar.entries();
		while (entries.hasMoreElements()) {
			// get a jar entry and input stream
			JarEntry entry = entries.nextElement();
			InputStream inJarStream = inJar.getInputStream(entry);

			// add the jar entry
			outJarStream.putNextEntry(new JarEntry(entry.getName()));
			if (entry.isDirectory()) {
				// do nothing for directory
			} else if (entry.getName().endsWith(".class")) {
				transformClassStream(inJarStream, outJarStream);
				// a jar file inside jar?
			}
			/*
			 * else if (entry.getName().endsWith(".jar")) { }
			 */
			else {
				Util.copyStream(inJarStream, outJarStream);
			}

			// close the inJar stream
			inJarStream.close();
		}

		outJarStream.close();
		inJar.close();
	}

	/**
	 * Transform a class file
	 * 
	 * @param classFile
	 * @param transformedClassFile
	 * @throws IOException
	 */
	private void transformClassFile(File classFile, File transformedClassFile)
			throws IOException {
		FileInputStream classFileInputStream = new FileInputStream(classFile);

		Logger.println("Instrumeting " + classFile + " --> "
				+ transformedClassFile);
		FileOutputStream fout = new FileOutputStream(transformedClassFile);
		transformClassStream(classFileInputStream, fout);
		classFileInputStream.close();
		fout.close();
	}

	/**
	 * Transform class stream
	 * 
	 * @param classFileInputStream
	 * @param fout
	 * @throws IOException
	 */
	private void transformClassStream(InputStream classFileInputStream,
			OutputStream fout) throws IOException {
		byte[] transformcledClassByte = treeAPITransform(classFileInputStream);
		fout.write(transformcledClassByte);
	}

	public byte[] treeAPITransform(InputStream classIn) throws IOException {
		ClassReader cr = new ClassReader(classIn);
		return treeAPITransform(cr);
	}

	public byte[] treeAPITransform(byte[] classIn) throws IOException {
		ClassReader cr = new ClassReader(classIn);
		return treeAPITransform(cr);
	}

	public byte[] treeAPITransform(ClassReader cr) throws IOException {
		ClassNode cn = new ClassNode();
		cr.accept(cn, ClassReader.SKIP_FRAMES);
		transformClassNode(cn);
		ClassWriter cw = new ClassWriter(ClassWriter.COMPUTE_MAXS);
		cn.accept(cw);

		if (false) {
			StringWriter sw = new StringWriter();
			PrintWriter pw = new PrintWriter(sw);
			CheckClassAdapter.verify(new ClassReader(cw.toByteArray()), false,
					pw);
			// TestCase.assertTrue(sw.toString(), sw.toString().length() == 0);
		}

		return cw.toByteArray();
	}

	/**
	 * The main class transformation method
	 * 
	 * @param cn
	 * @return
	 */
	@SuppressWarnings("unchecked")
	private boolean transformClassNode(@NonNull
	ClassNode cn) {
		// check if we need to instrument the given class
		if (!Util.shouldInstrumentThisClass(cn.name)) {
			return false;
		}

		// skip enum related ones
		if (cn.toString().indexOf("enum") != -1) {
			return false;
		}
		// No instrumentation for the enum type
		if ((cn.access & ACC_ENUM) != 0) {
			return false;
		}

		// No instrumentation for the interface
		if ((cn.access & ACC_INTERFACE) != 0) {
			return false;
		}

		// Skip inner classtransformClassNode
		if (false && cn.name.indexOf('$') != -1) {
			return false;
		}

		if (useClone) {
			addCloneableImplemetation(cn);
			myClonnableTypeSet.add(cn.name);
		}

		Hashtable<String, FieldNode> allFieldTypeTable = new Hashtable<String, FieldNode>();

		// ## Let's make a final filed a non final
		for (FieldNode fn : (List<FieldNode>) cn.fields) {
			fn.access &= ~ACC_FINAL;
			allFieldTypeTable.put(fn.name, fn);
		}

		boolean foundMainMethod = false;
		boolean foundCloneMethod = false;
		Type classNodeObjectType = Type.getObjectType(cn.name);

		// Walk through each method
		for (MethodNode mn : (List<MethodNode>) cn.methods) {
			TransformCounter.increaseMethod();

			String fullMethodName = Util.transClassNameSlashToDot(cn.name)
					+ "." + mn.name;

			// don't instrument the clone method if you are already in the class
			if (useClone && "clone".equals(mn.name)
					&& "()Ljava/lang/Object;".equals(mn.desc)) {
				Logger.info("We found clone for " + cn.name + "!");

				// Make sure it is public
				mn.access |= ACC_PUBLIC;
				foundCloneMethod = true;

				// Don't instrument the clone method
				continue;
			}

			// two-round mode
			if (scMode != null && !scMode.amIinTheBackStrackTrace(cn, mn)) {
				Logger.info("[second change mode] skip: " + fullMethodName);
				continue;
			}

			HashSet<String> staticObjectSet = new LinkedHashSet<String>();
			List<AbstractInsnNode> retInsertPointList = new ArrayList<AbstractInsnNode>();
			List<reCrashWritePoint> traceWriterInsertPointList = new ArrayList<reCrashWritePoint>();
			Hashtable<String, FieldNode> fieldUsedInThisMethodTable = new Hashtable<String, FieldNode>();

			// Check if this method is static void main.
			if ("main".equals(mn.name)
					&& "([Ljava/lang/String;)V".equals(mn.desc)
					&& (mn.access & ACC_PUBLIC) != 0
					&& (mn.access & ACC_STATIC) != 0) {
				Logger.info("We found main for " + cn.name + "!");

				// Add trace writer to record the program name and version #
				if (subjectProgramName != null) {
					InsnList il = new InsnList();

					il.add(new LdcInsnNode(subjectProgramName));
					il.add(new MethodInsnNode(INVOKESTATIC,
							"edu/mit/csail/pag/recrash/tracer/TraceWriter",
							"setProgramNameVersionNumber",
							"(Ljava/lang/String;)V"));

					mn.instructions.insert(il);
				}

				if (transformMain) {
					mn.name = Util.ORIGINAL_MAIN;
					foundMainMethod = true;
				}
			}

			if ("<init>".equals(mn.name) || "<clint>".equals(mn.name)) {
				continue;
			}

			if ((mn.access & ACC_ABSTRACT) != 0) { // no abstract
				continue;
			}

			if (false && (mn.access & ACC_STATIC) != 0
					&& !"main".equals(mn.name)) {
				// all static methods
				continue;
			}

			if ((mn.access & ACC_PUBLIC) == 0) { // no public
				continue;
			}

			// check instruction size and skip empty ones
			InsnList insns = mn.instructions;
			if (insns.size() == 0) {
				continue;
			}

			// get types from method description
			Type[] types = Type.getArgumentTypes(mn.desc);

			// Skip methods with no arguments
			if (types.length == 0) {
				continue;
			}

			// count # of instrumented methods
			TransformCounter.increaseInstrmentedmethod();
			Logger.info("Instrumenting " + mn.name + ":" + mn.desc);

			/*
			 * Static analysis part of the op code
			 */
			int opCount = 0;
			Iterator j = insns.iterator();

			while (j.hasNext()) {
				opCount++;
				AbstractInsnNode in = (AbstractInsnNode) j.next();

				int op = in.getOpcode();
				if (in instanceof LabelNode) {
					opCount--;
					continue; // do nothing with label
				}

				// Find all return ops and memorize the points
				// later we will instrument these points.
				if ((op >= IRETURN && op <= RETURN) /* || op == ATHROW */) {
					retInsertPointList.add(in.getPrevious());
					opCount--;
					continue;
				}

				/*
				 * We are looking for user annotated exception handling point
				 * ""reCrash with".equals(e);" LDC "reCrash with" ALOAD 2
				 * INVOKEVIRTUAL java/lang/String.equals(Ljava/lang/Object;)Z
				 */
				if (INVOKEVIRTUAL == op
						&& ((MethodInsnNode) in).name != null
						&& ((MethodInsnNode) in).name.equals("equals")
						&& ((MethodInsnNode) in).owner != null
						&& ((MethodInsnNode) in).owner
								.equals("java/lang/String")
						&& ((MethodInsnNode) in).desc != null
						&& ((MethodInsnNode) in).desc
								.equals("(Ljava/lang/Object;)Z")) {
					AbstractInsnNode prevIn = in.getPrevious();
					// look at the previous command
					if (prevIn != null && prevIn.getOpcode() == ALOAD) {
						AbstractInsnNode prevPrevIn = prevIn.getPrevious();

						if (prevPrevIn != null
								&& prevPrevIn.getOpcode() == LDC
								&& ((LdcInsnNode) prevPrevIn).cst != null
								&& ((LdcInsnNode) prevPrevIn).cst
										.equals(RECRASH_ANNOTATION)) {

							// add this point into the list. We will instrument
							// later
							traceWriterInsertPointList
									.add(new reCrashWritePoint(prevPrevIn
											.getPrevious(),
											((VarInsnNode) prevIn).var));
						}
					}
				}

				// TODO: need to figure out all statics
				// currently we ignore all static fields
				if (INVOKESTATIC == op) {
					MethodInsnNode mNode = (MethodInsnNode) in;
					if (Util.shouldInstrumentThisClass(mNode.owner)) {
						staticObjectSet.add(mNode.owner);
					}
				}

				if (PUTSTATIC == op || PUTSTATIC == op) {
					FieldInsnNode fNode = (FieldInsnNode) in;
					if (Util.shouldInstrumentThisClass(fNode.owner)) {
						staticObjectSet.add(fNode.owner);
					}
				}

				// find field used in this method
				if (GETFIELD == op || PUTFIELD == op) {
					FieldInsnNode fNode = (FieldInsnNode) in;

					// We take only my field
					if (fNode.owner.equals(cn.name)) {
						FieldNode fn = allFieldTypeTable.get(fNode.name);

						if (fn != null) { // ### FIXME: this only copies local
							// fields. No fields in the super
							Logger.info(fNode.owner + ":" + fNode.name + ":"
									+ fNode.desc);
							fieldUsedInThisMethodTable.put(fNode.name, fn);
						}
					}
				}

				// TODO: consider this call
				// INVOKEVIRTUAL
				// org/eclipse/jdt/internal/compiler/ast/Expression.resolveType(Lorg/eclipse/jdt/internal/compiler/lookup/BlockScope;)Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;
				// if a method calls this.method, the callee must include fields
				// in this.method as well
				if (INVOKEVIRTUAL == op) {
					MethodInsnNode mNode = (MethodInsnNode) in;
					if (cn.name.equals(mNode.owner)
							|| cn.superName.equals(mNode.owner)) {
						// fieldUsedInThisMethodTable = allFieldTypeTable;
					}
				}
			} // OPcode basic analysis is done

			// Empty block or getter/setter?
			if (opCount < minOpCount) {
				continue;
			}

			// Add the yourOK in all return points
			for (AbstractInsnNode retPoint : retInsertPointList) {
				InsnList il = new InsnList();
				il.add(new LdcInsnNode(fullMethodName));
				// use local variable
				il.add(new VarInsnNode(ILOAD, mn.maxLocals + 1));
				il.add(new MethodInsnNode(INVOKESTATIC,
						"edu/mit/csail/pag/recrash/tracer/TraceWriter",
						"youAreOK", "(Ljava/lang/String;I)V"));
				insns.insert(retPoint, il);
			}

			// Add trace writer for all annotated exception handling points
			for (reCrashWritePoint writePoint : traceWriterInsertPointList) {
				InsnList il = addCallMethodWriter(writePoint.exceptionArgNo);
				insns.insert(writePoint.location, il);
			}

			// Get new instruction list for argument capturing instrumentation
			InsnList il = new InsnList();

			boolean getReceiverOut = false;
			if (useClone && (mn.access & ACC_STATIC) == 0) {
				TransformCounter.totalClonedReceiver++;

				// If it is immutable
				if (mutability != null && mutability.isImmutable(mn, cn, "RC")) {
					TransformCounter.immutableObject++;

					// do nothing for the shallow clone
				} else { // if it is mutable
					getReceiverOut = true;
					il.add(new VarInsnNode(ALOAD, 0));
					String name = Util.transClassNameDotToSlash(cn.name);
					il.add(new MethodInsnNode(INVOKEVIRTUAL, name, "clone",
							"()Ljava/lang/Object;"));
					il.add(new TypeInsnNode(CHECKCAST, name));
					il.add(new VarInsnNode(ASTORE, mn.maxLocals));

					if (!useShallowClone) {
						Enumeration<String> em = fieldUsedInThisMethodTable
								.keys();
						while (em.hasMoreElements()) {
							String fieldName = em.nextElement();
							FieldNode fieldNode = fieldUsedInThisMethodTable
									.get(fieldName);

							try {
								cloneUsedFields(il, fieldNode, cn, mn);
							} catch (ClassNotFoundException e) {
								// TODO Auto-generated catch block
								e.printStackTrace();
							}
						}
					}
				}
			}

			il.add(new LdcInsnNode(fullMethodName));
			int arraySize = types.length + 1;
			// create array for objects
			// array length = arg num + 1 (for this)
			il.add(getInstForIConst(arraySize));
			il.add(new TypeInsnNode(ANEWARRAY, "java/lang/Object"));

			// keep local type size
			int localVarIndex = 0;

			// If it is not static
			if ((mn.access & ACC_STATIC) == 0) {

				if (getReceiverOut) { // declared the receiver out
					// add receiver in the object array
					il.add(new InsnNode(DUP));
					il.add(getInstForIConst(0));
					il.add(new VarInsnNode(ALOAD, mn.maxLocals));

					il.add(new InsnNode(AASTORE));
				} else {
					addObjectArrayElement(0, il, classNodeObjectType, 0, cn);
				}

				// increase the local var index if it is not static
				localVarIndex += classNodeObjectType.getSize();
			}

			for (int t = 0; t < types.length; t++) {
				Logger.info("Type object to instrument: " + types[t] + ": "
						+ types[t].getClassName());

				addObjectArrayElement(t + 1, il, types[t], localVarIndex, cn);
				// size of type can be different
				localVarIndex += types[t].getSize();
			}

			// create an array for types
			// array length = arg num + 1 (for this)
			il.add(getInstForIConst(arraySize));
			il.add(new TypeInsnNode(ANEWARRAY, "java/lang/String"));

			// add this type in the string array
			addTypeStringElement(0, il, classNodeObjectType);

			for (int t = 0; t < types.length; t++) {
				Logger.info("Type string to instrument: " + types[t] + ": "
						+ types[t].getClassName());
				addTypeStringElement(t + 1, il, types[t]);
			}

			// Add a boolean array which indicate if we need clone the object
			il.add(getInstForIConst(arraySize));
			il.add(new IntInsnNode(NEWARRAY, T_BOOLEAN));

			// add this type in the boolean array. Let's make it true for
			// now
			addBooleanToClone(0, il, classNodeObjectType);

			for (int t = 0; t < types.length; t++) {
				Logger.info(" string to instrument: " + types[t] + ": "
						+ types[t].getClassName());
				addBooleanToClone(t + 1, il, types[t]);
			}

			il
					.add(new MethodInsnNode(INVOKESTATIC,
							"edu/mit/csail/pag/recrash/tracer/TraceWriter",
							"youMayCrash",
							"(Ljava/lang/String;[Ljava/lang/Object;[Ljava/lang/String;[Z)I"));

			il.add(new VarInsnNode(ISTORE, mn.maxLocals + 1));

			// capture all add static objects (Not for now)
			if (false) {
				for (String staticClassName : staticObjectSet) {
					addStaticTraceCall(il, staticClassName);
				}
			}

			insns.insert(il);

			// Add local variable at the end
			LabelNode begin = new LabelNode();
			LabelNode end = new LabelNode();
			insns.insertBefore(insns.getFirst(), begin);
			insns.insert(insns.getLast(), end);

			mn.localVariables.add(new LocalVariableNode("__nounce_number__",
					"I", null, begin, end, mn.maxLocals + 1));

			mn.maxLocals += 2;
			// mn.maxStack += classNodeObjectType.getSize(); // for the local
		}

		if (transformMain && foundMainMethod) {
			cn.methods.add(addNewMain(cn));
		}

		// If it is a clone mode and does not have a clone method
		if (useClone && !foundCloneMethod) {
			// Add clone() method
			try {
				Class superClass = GetClass.fromTypeName(Util
						.transClassNameDotToSlash(cn.superName));
				Method method = superClass.getDeclaredMethod("clone");
				if (!Modifier.isFinal(method.getModifiers())) {
					cn.methods.add(addCloneMethod(cn));
				}
			} catch (SecurityException e) {
				// TODO Auto-generated catch block
				// e.printStackTrace();
			} catch (ClassNotFoundException e) {
				// TODO Auto-generated catch block
				// e.printStackTrace();
			} catch (NoSuchMethodException e) {
				cn.methods.add(addCloneMethod(cn));
				// e.printStackTrace();
			}

		}
		return foundMainMethod;
	}

	/**
	 * We clone used field in the corresponding method
	 * 
	 * @param il
	 * 
	 * @param fieldName
	 * @param fieldType
	 * @param cn
	 * @param mn
	 * @throws ClassNotFoundException
	 */
	private boolean cloneUsedFields(InsnList il, FieldNode fieldNode,
			ClassNode cn, MethodNode mn) throws ClassNotFoundException {
		String fieldName = fieldNode.name;
		String fieldType = fieldNode.desc;

		Logger.info(">>>" + fieldName + ":" + fieldType);

		// array
		if (fieldType.indexOf('[') != -1) {
			return false;
		}

		// will take care of by clone method, shallow copy
		if (GetClass.isPrimitive(fieldType)) {
			return false;
		}

		// will take care of by clone method, shallow copy
		if (GetClass.isPrimitiveShortType(fieldType)) {
			return false;
		}
		String slashedTypeName = Util.transClassNameDotToSlash(fieldType);
		String shortTypeName = slashedTypeName.substring(1, slashedTypeName
				.length() - 1);

		String slashedClassTypeName = Util.transClassNameDotToSlash(cn.name);

		if (GetClass.isPrimitiveJavaLangType(fieldType)) {
			return false;
		}

		if ("java/lang/String".equals(slashedTypeName)) {
			return false;
		}

		// No static clone: It's a static anyway
		if ((fieldNode.access & ACC_STATIC) != 0) {
			return false;
		}

		// TODO: it is known as immutable in the method we should return

		// TODO: if it is not our object, we may skip
		if (!myClonnableTypeSet.contains(shortTypeName)) {
			Logger.info("!!! No class in my cloneable list: " + shortTypeName);
			return false;
		}

		Logger.info("<<< doing clone" + fieldName + ":" + fieldType + "("
				+ shortTypeName + ")");

		// Null test
		LabelNode L = new LabelNode();
		il.add(new VarInsnNode(ALOAD, 0));
		il.add(new FieldInsnNode(GETFIELD, slashedClassTypeName, fieldName,
				fieldType));
		il.add(new JumpInsnNode(IFNULL, L));

		// load new cloned object
		il.add(new VarInsnNode(ALOAD, mn.maxLocals));

		// load this
		il.add(new VarInsnNode(ALOAD, 0));

		il.add(new FieldInsnNode(GETFIELD, slashedClassTypeName, fieldName,
				fieldType));
		il.add(new MethodInsnNode(INVOKEVIRTUAL, shortTypeName, "clone",
				"()Ljava/lang/Object;"));
		il.add(new TypeInsnNode(CHECKCAST, shortTypeName));
		il.add(new FieldInsnNode(PUTFIELD, slashedClassTypeName, fieldName,
				fieldType));
		il.add(L);

		TransformCounter.totalClonedField++;
		return true;
	}

	/**
	 * Add an boolean element for clone indicator
	 * 
	 * @param index
	 * @param il
	 * @param cn
	 * @param mn
	 */
	private void addBooleanToClone(int index, InsnList il, Type type) {

		// Check if we need to clone this i-th argument in the method/class
		// it is added in the instrumentation level
		if (useSerialization && needToBeClone(index, type)) {
			il.add(new InsnNode(DUP));
			il.add(getInstForIConst(index));

			il.add(new InsnNode(ICONST_1));
			il.add(new InsnNode(BASTORE));
		}
	}

	@SuppressWarnings("unchecked")
	private void addCloneableImplemetation(ClassNode cn) {
		List implementList = cn.interfaces;
		if (!implementList.contains("java/lang/Cloneable")) {
			implementList.add("java/lang/Cloneable");
		}
	}

	/**
	 * 
	 * @param iConst
	 * @return
	 */
	private AbstractInsnNode getInstForIConst(int iConst) {
		switch (iConst) {
		case 0:
			return new InsnNode(ICONST_0);
		case 1:
			return new InsnNode(ICONST_1);
		case 2:
			return new InsnNode(ICONST_2);
		case 3:
			return new InsnNode(ICONST_3);
		case 4:
			return new InsnNode(ICONST_4);
		case 5:
			return new InsnNode(ICONST_5);
		default:
			return new IntInsnNode(BIPUSH, iConst);
		}
	}

	/**
	 * Add one static object trace
	 * 
	 * @param il
	 * @param staticClassName
	 *            full static object name
	 */
	private void addStaticTraceCall(InsnList il, String staticClassName) {
		il.add(new LdcInsnNode(staticClassName));
		il.add(new MethodInsnNode(INVOKESTATIC,
				"edu/mit/csail/pag/recrash/tracer/TraceWriter", "addStatic",
				"(Ljava/lang/String;)V"));
	}

	@SuppressWarnings("unchecked")
	private MethodNode addNewMain(ClassNode cn) {
		MethodNode newMain = new MethodNode(ACC_PUBLIC + ACC_STATIC, "main",
				"([Ljava/lang/String;)V", null,
				new String[] { "java/lang/Throwable" });
		InsnList il = newMain.instructions;

		List localVariables = newMain.localVariables;
		List tryCatchBolcks = newMain.tryCatchBlocks;

		LabelNode L0 = new LabelNode();
		LabelNode L1 = new LabelNode();
		LabelNode L2 = new LabelNode();
		LabelNode L3 = new LabelNode();
		LabelNode L4 = new LabelNode();
		LabelNode L5 = new LabelNode();
		LabelNode L6 = new LabelNode();
		LabelNode L7 = new LabelNode();

		tryCatchBolcks.add(new TryCatchBlockNode(L0, L1, L2,
				"java/lang/Throwable"));

		il.add(L0);
		il.add(new VarInsnNode(ALOAD, 0));
		il.add(new MethodInsnNode(INVOKESTATIC, cn.name, Util.ORIGINAL_MAIN,
				"([Ljava/lang/String;)V"));
		il.add(L1);
		il.add(new JumpInsnNode(GOTO, L3));
		il.add(L2);
		il.add(new VarInsnNode(ASTORE, 1));
		il.add(L4);

		// add method call writer
		il.add(addCallMethodWriter(1));

		il.add(L5);

		// throw the exception
		il.add(new VarInsnNode(ALOAD, 1));
		il.add(new InsnNode(ATHROW));
		il.add(L3);

		il.add(L6);
		il.add(new InsnNode(RETURN));
		il.add(L7);
		localVariables.add(new LocalVariableNode("args", "[Ljava/lang/String;",
				null, L0, L7, 0));

		// localVariables.add(new LocalVariableNode("e",
		// "Ljava/lang/Throwable;",
		localVariables.add(new LocalVariableNode("e", "Ljava/lang/Throwable;",
				null, L4, L3, 1));

		newMain.maxStack = 1;
		newMain.maxLocals = 2;

		return newMain;
	}

	private InsnList addCallMethodWriter(int exceptionArgNumber) {
		InsnList il = new InsnList();
		il.add(new VarInsnNode(ALOAD, exceptionArgNumber));
		il.add(new MethodInsnNode(INVOKESTATIC,
				"edu/mit/csail/pag/recrash/tracer/TraceWriter", "writeTrace",
				"(Ljava/lang/Throwable;)Ljava/io/File;"));
		il.add(new InsnNode(POP));

		return il;
	}

	@SuppressWarnings("unchecked")
	MethodNode addCloneMethod(ClassNode cn) {
		MethodNode cloneMethod = new MethodNode(ACC_PUBLIC, "clone",
				"()Ljava/lang/Object;", null, null);

		InsnList il = cloneMethod.instructions;

		List tryCatchBolcks = cloneMethod.tryCatchBlocks;

		LabelNode L0 = new LabelNode();
		LabelNode L1 = new LabelNode();
		LabelNode L2 = new LabelNode();
		LabelNode L3 = new LabelNode();
		LabelNode L4 = new LabelNode();
		LabelNode L5 = new LabelNode();

		tryCatchBolcks.add(new TryCatchBlockNode(L0, L1, L2,
				"java/lang/Throwable"));

		il.add(L0);
		il.add(new VarInsnNode(ALOAD, 0));
		il.add(new MethodInsnNode(INVOKESPECIAL, cn.superName, "clone",
				"()Ljava/lang/Object;"));
		il.add(new InsnNode(ARETURN));
		il.add(L1);
		// il.add(new InsnNode(ARETURN));
		il.add(L2);
		il.add(new VarInsnNode(ASTORE, 1));
		il.add(L3);
		il.add(new VarInsnNode(ALOAD, 1));
		il.add(new MethodInsnNode(INVOKEVIRTUAL, "java/lang/Throwable",
				"printStackTrace", "()V"));
		il.add(L4);
		il.add(new VarInsnNode(ALOAD, 0));
		il.add(new InsnNode(ARETURN));
		il.add(L5);
		cloneMethod.maxStack = 1;
		cloneMethod.maxLocals = 2;
		return cloneMethod;
	}

	private void addTypeStringElement(int index, InsnList il, Type type) {
		il.add(new InsnNode(DUP));
		il.add(getInstForIConst(index));

		il.add(new LdcInsnNode(type.getClassName()));
		il.add(new InsnNode(AASTORE));
	}

	/**
	 * Make array of arguments
	 * @param index
	 * @param il
	 * @param type
	 * @param argLocation
	 * @param cn
	 */
	private void addObjectArrayElement(int index, InsnList il, Type type,
			int argLocation, ClassNode cn) {
		il.add(new InsnNode(DUP));
		il.add(getInstForIConst(index));

		if (type.equals(Type.BOOLEAN_TYPE)) {
			il.add(new VarInsnNode(ILOAD, argLocation));
			il.add(new MethodInsnNode(INVOKESTATIC, "java/lang/Boolean",
					"valueOf", "(Z)Ljava/lang/Boolean;"));
			// typeClassName = Boolean.class.getCanonicalName();
		} else if (type.equals(Type.CHAR_TYPE)) {
			il.add(new VarInsnNode(ILOAD, argLocation));
			il.add(new MethodInsnNode(INVOKESTATIC, "java/lang/Character",
					"valueOf", "(C)Ljava/lang/Character;"));
			// typeClassName = Character.class.getCanonicalName();
		} else if (type.equals(Type.BYTE_TYPE)) {
			il.add(new VarInsnNode(ILOAD, argLocation));
			il.add(new MethodInsnNode(INVOKESTATIC, "java/lang/Byte",
					"valueOf", "(B)Ljava/lang/Byte;"));
			// typeClassName = Boolean.class.getCanonicalName();
		} else if (type.equals(Type.SHORT_TYPE)) {
			il.add(new VarInsnNode(ILOAD, argLocation));
			il.add(new MethodInsnNode(INVOKESTATIC, "java/lang/Short",
					"valueOf", "(S)Ljava/lang/Short;"));
			// typeClassName = Short.class.getCanonicalName();
		} else if (type.equals(Type.INT_TYPE)) {
			il.add(new VarInsnNode(ILOAD, argLocation));
			il.add(new MethodInsnNode(INVOKESTATIC, "java/lang/Integer",
					"valueOf", "(I)Ljava/lang/Integer;"));
			// typeClassName = Integer.class.getCanonicalName();
		} else if (type.equals(Type.FLOAT_TYPE)) {
			il.add(new VarInsnNode(FLOAD, argLocation));
			il.add(new MethodInsnNode(INVOKESTATIC, "java/lang/Float",
					"valueOf", "(F)Ljava/lang/Float;"));
			// typeClassName = Float.class.getCanonicalName();
		} else if (type.equals(Type.LONG_TYPE)) {
			il.add(new VarInsnNode(LLOAD, argLocation));
			il.add(new MethodInsnNode(INVOKESTATIC, "java/lang/Long",
					"valueOf", "(J)Ljava/lang/Long;"));
			// typeClassName = Long.class.getCanonicalName();
		} else if (type.equals(Type.DOUBLE_TYPE)) {
			il.add(new VarInsnNode(DLOAD, argLocation));
			il.add(new MethodInsnNode(INVOKESTATIC, "java/lang/Double",
					"valueOf", "(D)Ljava/lang/Double;"));
			// typeClassName = Double.class.getCanonicalName();
		} else {
			il.add(new VarInsnNode(ALOAD, argLocation));

			// pass cloned object
			// It will be done when write object in Trace Writer.
			if (false) {
				System.out.println("Add args: " + useClone + ":"
						+ needToBeClone(index, type));
				if (useClone && needToBeClone(index, type)) {
					String name = Util.transClassNameDotToSlash(type
							.getClassName());
					System.out.println("Add cloned args");
					if (Util.shouldInstrumentThisClass(name)) {
						il.add(new MethodInsnNode(INVOKEVIRTUAL, name, "clone",
								"()Ljava/lang/Object;"));
					}
				}
			}
		}

		il.add(new InsnNode(AASTORE));
	}

	/**
	 * TODO: Need to check condition to add clone
	 * @param index
	 * @param type
	 * @return
	 */
	private boolean needToBeClone(int index, Type type) {
		if (index == 0) { // no need to clone again for receiver
			return false;
		}

		// TODO: we need to make sure that the clone method is accessible
		// We need to have the subject class in reCrash's classpath to figure it out
		if (false) {
			try {
				Class<?> clazz = GetClass.fromTypeName(type.getClassName());
				Method method = clazz.getDeclaredMethod("clone");
				method.setAccessible(true);
			} catch (Exception e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
				return false;
			}
		}

		return true;
	}

	public static void main(String args[]) throws IOException {
		if (args.length < 2) {
			usage();
			return;
		}

		TransformCounter.init();

		Transformer trans = new Transformer();

		trans.parseArgs(args, 2);

		trans.transformDir(new File(args[0]), new File(args[1]));

		if (trans.useClone) { // we need two path transformation
			TransformCounter.init();
			trans.transformDir(new File(args[0]), new File(args[1]));
		}

		Logger.info(TransformCounter.report());
	}

	public void parseArgs(String[] args, int argStartNo) throws IOException {
		// default value
		transformMain = true;
		useClone = false;

		for (int i = argStartNo; i < args.length; i++) {
			if (args[i].equals("-verbose")) {
				Logger.verbose = true;
			} else if (args[i].equals("-skip_main")) {
				Logger.info("Skip main method!");
				transformMain = false;
			} else if (args[i].equals("-pname")) {
				subjectProgramName = args[++i];
			} else if (args[i].equals("-mut_data")) {
				File dataFile = new File(args[++i]);
				if (!dataFile.exists()) {
					Logger.error("No mutability data file: "
							+ dataFile.getAbsolutePath());
					return;
				}
				mutability = new Mutability(dataFile.getAbsolutePath());
			} else if (args[i].equals("-second_chance")) {
				File dataFile = new File(args[++i]);
				if (!dataFile.exists()) {
					Logger.error("No second chance data file: "
							+ dataFile.getAbsolutePath());
					return;
				}
				scMode = new SecondChangeMode(dataFile.getAbsolutePath());
			} else if (args[i].equals("-use_clone")) {
				useClone = true;
			} else if (args[i].equals("-use_shallow_clone")) {
				useClone = true;
				useShallowClone = true;
			} else if (args[i].equals("-use_serialization")) {
				useSerialization = true;
			} else {
				System.err.println("What?? " + args[i]);
				usage();
				return;
			}
		}
	}

	/**
	 * 
	 */
	public static void usage() {
		System.out.println("transform <src> <desc> <options>");
		System.out.println("Options:");
		System.out.println("  -verbose (prints information)");
		System.out.println("  -skip_main (do not replae main)");
		System.out
				.println("  -use_clone (Clone the receiver and all used fields using Object.clone()");
		System.out
				.println("  -use_shallow_clone (Clone only the receiver using Object.clone())");
		System.out.println("  -use_serialization (use serialization)");
		System.out
				.println("  -pname <program name> (Program name to display in test cases)");
		System.out.println("  -mut_data <mutability info> (Mutability info)");
		System.out
				.println("  -second_chance <second chance data file> (Stack information)");
		System.exit(0);
	}
}
