import edu.mit.csail.pag.recrash.GetClass;
import junit.framework.TestCase;

public class GetClassTest extends TestCase {

	public void test_fromTypeName() throws ClassNotFoundException {
		Class<?> clazz = GetClass.fromTypeName("java.lang.Integer");
		assertEquals(clazz, (new Integer(10)).getClass());

		clazz = GetClass.fromTypeName("int");
		assertEquals(clazz, int.class);

		clazz = GetClass.fromTypeName("int[]");
		assertEquals(clazz, int[].class);
	}

	public void test_descToSig() throws Exception {
		String descs[] = {
				"Ljava/lang/Integer;I",
				"()Z",
				"(I)Ljava/lang/String;",
				"()Ljava/lang/String;",
				"(Ljava/lang/String;Z)V",
				"()[Lorg/tmatesoft/svn/core/io/SVNError;",
				"(Ljava/lang/String;)Lorg/tmatesoft/svn/core/ISVNEntryContent;",
				"(Ljava/lang/String;Ljava/util/Collection;JJ)Ljava/util/Collection;",
				"(Lorg/tmatesoft/svn/core/io/ISVNCredentials;Ljava/lang/String;)V",
				"([Ljava/lang/String;Ljava/util/Collection;JJZZ)Ljava/util/Collection;" };
		String expacts[] = { "java/lang/Integer,int", "", "int", "",
				"java/lang/String,boolean", "", "java/lang/String",
				"java/util/Collection,long,long", "java/lang/String",
				"java/util/Collection,long,long,boolean,boolean" };

		assertEquals(descs.length, expacts.length);

		for (int i = 0; i < descs.length; i++) {
			assertEquals("desctoSig " + i, expacts[i], GetClass
					.descToSig(descs[i]));
		}
	}

	public void test_getCorrespondingObjectName() {
		String names[] = { "int", "void", "double", "java/lang/String" };
		String expacts[] = { "Integer", "Void", "Double", "java/lang/String" };
		assertEquals(names.length, expacts.length);

		for (int i = 0; i < names.length; i++) {
			assertEquals("getCorrespondingObjectName " + i, expacts[i],
					GetClass.getCorrespondingObjectName(names[i]));
		}
	}

	public void test_isPrimitive() throws ClassNotFoundException {
		String names[] = { "int", "void", "double[]", "java/lang/String" };
		boolean expacts[] = { true, true, false, false };
		assertEquals(names.length, expacts.length);

		for (int i = 0; i < names.length; i++) {
			assertEquals("isPrimitive " + i, expacts[i], GetClass
					.isPrimitive(names[i]));
		}
	}

	public void test_isJavaPrimitive() throws ClassNotFoundException {
		String names[] = { "int", "void", "double[]", "java/lang/String",
				"Ljava/lang/Integer;", "java.lang.Void" };
		boolean expacts[] = { false, false, false, false, false, true };
		assertEquals(names.length, expacts.length);

		for (int i = 0; i < names.length; i++) {
			assertEquals("sPrimitiveJavaLangType " + i, expacts[i], GetClass
					.isPrimitiveJavaLangType(names[i]));
		}
	}
}
